@@ -71,17 +71,22 @@
 * Includes.
 \******************************************************************************/
 
+#include "jp2_dec.h"
+#include "jp2_cod.h"
+
 #include "jasper/jas_image.h"
 #include "jasper/jas_stream.h"
 #include "jasper/jas_math.h"
 #include "jasper/jas_debug.h"
 #include "jasper/jas_malloc.h"
-#include "jasper/jas_version.h"
+#include "jasper/jas_types.h"
 
-#include "jp2_cod.h"
-#include "jp2_dec.h"
+#include <assert.h>
+#include <stdio.h>
 
-#define	JP2_VALIDATELEN	(JAS_MIN(JP2_JP_LEN + 16, JAS_STREAM_MAXPUTBACK))
+/******************************************************************************\
+* Function prototypes.
+\******************************************************************************/
 
 static jp2_dec_t *jp2_dec_create(void);
 static void jp2_dec_destroy(jp2_dec_t *dec);
@@ -93,7 +98,7 @@ static int jp2_getct(int colorspace, int type, int assoc);
 * Functions.
 \******************************************************************************/
 
-jas_image_t *jp2_decode(jas_stream_t *in, char *optstr)
+jas_image_t *jp2_decode(jas_stream_t *in, const char *optstr)
 {
 	jp2_box_t *box;
 	int found;
@@ -104,7 +109,10 @@ jas_image_t *jp2_decode(jas_stream_t *in, char *optstr)
 	unsigned int i;
 	jp2_cmap_t *cmapd;
 	jp2_pclr_t *pclrd;
+#if 0
+	/* The following code appears to no longer be needed. */
 	jp2_cdef_t *cdefd;
+#endif
 	unsigned int channo;
 	int newcmptno;
 	int_fast32_t *lutents;
@@ -120,21 +128,25 @@ jas_image_t *jp2_decode(jas_stream_t *in, char *optstr)
 	box = 0;
 	image = 0;
 
+	JAS_LOGDEBUGF(100, "jp2_decode(%p, \"%s\")\n", in, optstr);
+
 	if (!(dec = jp2_dec_create())) {
 		goto error;
 	}
 
 	/* Get the first box.  This should be a JP box. */
 	if (!(box = jp2_box_get(in))) {
-		jas_eprintf("error: cannot get box\n");
+		jas_logerrorf("error: cannot get box\n");
 		goto error;
 	}
 	if (box->type != JP2_BOX_JP) {
-		jas_eprintf("error: expecting signature box\n");
+		jas_logerrorf("error: expecting signature box\n");
 		goto error;
 	}
 	if (box->data.jp.magic != JP2_JP_MAGIC) {
-		jas_eprintf("incorrect magic number\n");
+		jas_logerrorf("incorrect magic number (0x%lx != 0x%lx)\n",
+		  JAS_CAST(unsigned long, box->data.jp.magic),
+		  JAS_CAST(unsigned long, JP2_JP_MAGIC));
 		goto error;
 	}
 	jp2_box_destroy(box);
@@ -142,10 +154,11 @@ jas_image_t *jp2_decode(jas_stream_t *in, char *optstr)
 
 	/* Get the second box.  This should be a FTYP box. */
 	if (!(box = jp2_box_get(in))) {
+		jas_logerrorf("error: cannot get box\n");
 		goto error;
 	}
 	if (box->type != JP2_BOX_FTYP) {
-		jas_eprintf("expecting file type box\n");
+		jas_logerrorf("expecting file type box\n");
 		goto error;
 	}
 	jp2_box_destroy(box);
@@ -154,8 +167,8 @@ jas_image_t *jp2_decode(jas_stream_t *in, char *optstr)
 	/* Get more boxes... */
 	found = 0;
 	while ((box = jp2_box_get(in))) {
-		if (jas_getdbglevel() >= 1) {
-			jas_eprintf("box type %s\n", box->info->name);
+		if (jas_get_debug_level() >= 1) {
+			jas_logdebugf(1, "got box type %s\n", box->info->name);
 		}
 		switch (box->type) {
 		case JP2_BOX_JP2C:
@@ -208,37 +221,39 @@ jas_image_t *jp2_decode(jas_stream_t *in, char *optstr)
 	}
 
 	if (!found) {
-		jas_eprintf("error: no code stream found\n");
+		jas_logerrorf("error: no code stream found\n");
 		goto error;
 	}
 
 	if (!(dec->image = jpc_decode(in, optstr))) {
-		jas_eprintf("error: cannot decode code stream\n");
+		jas_logerrorf("error: cannot decode code stream\n");
 		goto error;
 	}
 
 	/* An IHDR box must be present. */
 	if (!dec->ihdr) {
-		jas_eprintf("error: missing IHDR box\n");
+		jas_logerrorf("error: missing IHDR box\n");
 		goto error;
 	}
 
 	/* Does the number of components indicated in the IHDR box match
 	  the value specified in the code stream? */
-	if (dec->ihdr->data.ihdr.numcmpts != JAS_CAST(uint, jas_image_numcmpts(dec->image))) {
-		jas_eprintf("warning: number of components mismatch\n");
+	if (dec->ihdr->data.ihdr.numcmpts != JAS_CAST(jas_uint,
+	  jas_image_numcmpts(dec->image))) {
+		jas_logerrorf("error: number of components mismatch (IHDR)\n");
+		goto error;
 	}
 
 	/* At least one component must be present. */
 	if (!jas_image_numcmpts(dec->image)) {
-		jas_eprintf("error: no components\n");
+		jas_logerrorf("error: no components\n");
 		goto error;
 	}
 
 	/* Determine if all components have the same data type. */
 	samedtype = true;
 	dtype = jas_image_cmptdtype(dec->image, 0);
-	for (i = 1; i < JAS_CAST(uint, jas_image_numcmpts(dec->image)); ++i) {
+	for (i = 1; i < JAS_CAST(jas_uint, jas_image_numcmpts(dec->image)); ++i) {
 		if (jas_image_cmptdtype(dec->image, i) != dtype) {
 			samedtype = false;
 			break;
@@ -249,38 +264,43 @@ jas_image_t *jp2_decode(jas_stream_t *in, char *optstr)
 	  with the data in the code stream? */
 	if ((samedtype && dec->ihdr->data.ihdr.bpc != JP2_DTYPETOBPC(dtype)) ||
 	  (!samedtype && dec->ihdr->data.ihdr.bpc != JP2_IHDR_BPCNULL)) {
-		jas_eprintf("warning: component data type mismatch\n");
+		jas_logerrorf("error: component data type mismatch (IHDR)\n");
+		goto error;
 	}
 
 	/* Is the compression type supported? */
 	if (dec->ihdr->data.ihdr.comptype != JP2_IHDR_COMPTYPE) {
-		jas_eprintf("error: unsupported compression type\n");
+		jas_logerrorf("error: unsupported compression type\n");
 		goto error;
 	}
 
 	if (dec->bpcc) {
 		/* Is the number of components indicated in the BPCC box
 		  consistent with the code stream data? */
-		if (dec->bpcc->data.bpcc.numcmpts != JAS_CAST(uint, jas_image_numcmpts(
-		  dec->image))) {
-			jas_eprintf("warning: number of components mismatch\n");
+		if (dec->bpcc->data.bpcc.numcmpts !=
+		  JAS_CAST(jas_uint, jas_image_numcmpts(dec->image))) {
+			jas_logerrorf("error: number of components mismatch (BPCC)\n");
+			goto error;
 		}
 		/* Is the component data type information indicated in the BPCC
 		  box consistent with the code stream data? */
 		if (!samedtype) {
-			for (i = 0; i < JAS_CAST(uint, jas_image_numcmpts(dec->image)); ++i) {
-				if (jas_image_cmptdtype(dec->image, i) != JP2_BPCTODTYPE(dec->bpcc->data.bpcc.bpcs[i])) {
-					jas_eprintf("warning: component data type mismatch\n");
+			for (i = 0; i < JAS_CAST(jas_uint, jas_image_numcmpts(dec->image));
+			  ++i) {
+				if (jas_image_cmptdtype(dec->image, i) !=
+				  JP2_BPCTODTYPE(dec->bpcc->data.bpcc.bpcs[i])) {
+					jas_logerrorf("error: component data type mismatch (BPCC)\n");
+						goto error;
 				}
 			}
 		} else {
-			jas_eprintf("warning: superfluous BPCC box\n");
+			jas_logwarnf("warning: superfluous BPCC box\n");
 		}
 	}
 
 	/* A COLR box must be present. */
 	if (!dec->colr) {
-		jas_eprintf("error: no COLR box\n");
+		jas_logerrorf("error: no COLR box\n");
 		goto error;
 	}
 
@@ -292,55 +312,63 @@ jas_image_t *jp2_decode(jas_stream_t *in, char *optstr)
 		iccprof = jas_iccprof_createfrombuf(dec->colr->data.colr.iccp,
 		  dec->colr->data.colr.iccplen);
 		if (!iccprof) {
-			jas_eprintf("error: failed to parse ICC profile\n");
+			jas_logerrorf("error: failed to parse ICC profile\n");
 			goto error;
 		}
 		jas_iccprof_gethdr(iccprof, &icchdr);
-		jas_eprintf("ICC Profile CS %08x\n", icchdr.colorspc);
-		jas_image_setclrspc(dec->image, fromiccpcs(icchdr.colorspc));
+		jas_loginfof("ICC Profile CS %08x\n", icchdr.colorspc);
 		dec->image->cmprof_ = jas_cmprof_createfromiccprof(iccprof);
-		assert(dec->image->cmprof_);
+		if (!dec->image->cmprof_) {
+			jas_image_setclrspc(dec->image, JAS_CLRSPC_UNKNOWN);
+			jas_logwarnf("warning: cannot create CM profile from ICC profile\n");
+		} else {
+			jas_image_setclrspc(dec->image, fromiccpcs(icchdr.colorspc));
+		}
 		jas_iccprof_destroy(iccprof);
 		break;
 	}
 
 	/* If a CMAP box is present, a PCLR box must also be present. */
 	if (dec->cmap && !dec->pclr) {
-		jas_eprintf("warning: missing PCLR box or superfluous CMAP box\n");
+		jas_logwarnf("warning: missing PCLR box or superfluous CMAP box\n");
 		jp2_box_destroy(dec->cmap);
 		dec->cmap = 0;
 	}
 
 	/* If a CMAP box is not present, a PCLR box must not be present. */
 	if (!dec->cmap && dec->pclr) {
-		jas_eprintf("warning: missing CMAP box or superfluous PCLR box\n");
+		jas_logwarnf("warning: missing CMAP box or superfluous PCLR box\n");
 		jp2_box_destroy(dec->pclr);
 		dec->pclr = 0;
 	}
 
 	/* Determine the number of channels (which is essentially the number
 	  of components after any palette mappings have been applied). */
-	dec->numchans = dec->cmap ? dec->cmap->data.cmap.numchans : JAS_CAST(uint, jas_image_numcmpts(dec->image));
+	dec->numchans = dec->cmap ? dec->cmap->data.cmap.numchans :
+	  JAS_CAST(jas_uint, jas_image_numcmpts(dec->image));
 
 	/* Perform a basic sanity check on the CMAP box if present. */
 	if (dec->cmap) {
 		for (i = 0; i < dec->numchans; ++i) {
 			/* Is the component number reasonable? */
-			if (dec->cmap->data.cmap.ents[i].cmptno >= JAS_CAST(uint, jas_image_numcmpts(dec->image))) {
-				jas_eprintf("error: invalid component number in CMAP box\n");
+			if (dec->cmap->data.cmap.ents[i].cmptno >= JAS_CAST(jas_uint,
+			  jas_image_numcmpts(dec->image))) {
+				jas_logerrorf("error: invalid component number in CMAP box\n");
 				goto error;
 			}
 			/* Is the LUT index reasonable? */
-			if (dec->cmap->data.cmap.ents[i].pcol >= dec->pclr->data.pclr.numchans) {
-				jas_eprintf("error: invalid CMAP LUT index\n");
+			if (dec->cmap->data.cmap.ents[i].pcol >=
+			  dec->pclr->data.pclr.numchans) {
+				jas_logerrorf("error: invalid CMAP LUT index\n");
 				goto error;
 			}
 		}
 	}
 
 	/* Allocate space for the channel-number to component-number LUT. */
-	if (!(dec->chantocmptlut = jas_malloc(dec->numchans * sizeof(uint_fast16_t)))) {
-		jas_eprintf("error: no memory\n");
+	if (!(dec->chantocmptlut = jas_alloc2(dec->numchans,
+	  sizeof(uint_fast16_t)))) {
+		jas_logerrorf("error: no memory\n");
 		goto error;
 	}
 
@@ -349,23 +377,48 @@ jas_image_t *jp2_decode(jas_stream_t *in, char *optstr)
 			dec->chantocmptlut[i] = i;
 		}
 	} else {
+		/* Check to ensure that CMAP/PCLR were initialized. */
+		if (!dec->cmap || !dec->pclr) {
+			jas_logerrorf("missing CMAP/PCLR box\n");
+			goto error;
+		}
+
 		cmapd = &dec->cmap->data.cmap;
 		pclrd = &dec->pclr->data.pclr;
+#if 0
+		/* The following code appears to no longer be needed. */
 		cdefd = &dec->cdef->data.cdef;
+#endif
 		for (channo = 0; channo < cmapd->numchans; ++channo) {
 			cmapent = &cmapd->ents[channo];
 			if (cmapent->map == JP2_CMAP_DIRECT) {
 				dec->chantocmptlut[channo] = channo;
 			} else if (cmapent->map == JP2_CMAP_PALETTE) {
-				lutents = jas_malloc(pclrd->numlutents * sizeof(int_fast32_t));
+				if (!pclrd->numlutents) {
+					jas_logerrorf("error: empty LUT\n");
+					goto error;
+				}
+				if (!(lutents = jas_alloc2(pclrd->numlutents,
+				  sizeof(int_fast32_t)))) {
+					jas_logerrorf("error: cannot allocate LUT\n");
+					goto error;
+				}
 				for (i = 0; i < pclrd->numlutents; ++i) {
 					lutents[i] = pclrd->lutdata[cmapent->pcol + i * pclrd->numchans];
 				}
 				newcmptno = jas_image_numcmpts(dec->image);
-				jas_image_depalettize(dec->image, cmapent->cmptno, pclrd->numlutents, lutents, JP2_BPCTODTYPE(pclrd->bpc[cmapent->pcol]), newcmptno);
+				if (jas_image_depalettize(dec->image, cmapent->cmptno,
+				  pclrd->numlutents, lutents,
+				  JP2_BPCTODTYPE(pclrd->bpc[cmapent->pcol]), newcmptno)) {
+					jas_logerrorf("jas_image_depalettize failed\n");
+					jas_free(lutents);
+					jas_logerrorf("error: depalettize failed\n");
+					goto error;
+				}
 				dec->chantocmptlut[channo] = newcmptno;
 				jas_free(lutents);
 #if 0
+				/* The following code appears to no longer be needed. */
 				if (dec->cdef) {
 					cdefent = jp2_cdef_lookup(cdefd, channo);
 					if (!cdefent) {
@@ -376,32 +429,61 @@ jas_image_t *jp2_decode(jas_stream_t *in, char *optstr)
 				jas_image_setcmpttype(dec->image, newcmptno, jp2_getct(jas_image_clrspc(dec->image), 0, channo + 1));
 				}
 #endif
+			} else {
+				jas_logerrorf("error: invalid MTYP in CMAP box\n");
+				goto error;
 			}
 		}
 	}
 
+#if 0
+	/* The following code appears to no longer be needed. */
+	/* Ensure that the number of channels being used by the decoder
+	  matches the number of image components. */
+	if (dec->numchans != jas_image_numcmpts(dec->image)) {
+		jas_logerrorf("error: mismatch in number of components (%d != %d)\n",
+		  dec->numchans, jas_image_numcmpts(dec->image));
+		goto error;
+	}
+#endif
+
 	/* Mark all components as being of unknown type. */
 
-	for (i = 0; i < JAS_CAST(uint, jas_image_numcmpts(dec->image)); ++i) {
+	for (i = 0; i < JAS_CAST(jas_uint, jas_image_numcmpts(dec->image)); ++i) {
 		jas_image_setcmpttype(dec->image, i, JAS_IMAGE_CT_UNKNOWN);
 	}
 
 	/* Determine the type of each component. */
 	if (dec->cdef) {
-		for (i = 0; i < dec->numchans; ++i) {
+		for (i = 0; i < dec->cdef->data.cdef.numchans; ++i) {
+			uint_fast16_t channo = dec->cdef->data.cdef.ents[i].channo;
 			/* Is the channel number reasonable? */
-			if (dec->cdef->data.cdef.ents[i].channo >= dec->numchans) {
-				jas_eprintf("error: invalid channel number in CDEF box\n");
+			if (channo >= dec->numchans) {
+				jas_logerrorf("error: invalid channel number in CDEF box (%d)\n",
+				  channo);
 				goto error;
 			}
-			jas_image_setcmpttype(dec->image,
-			  dec->chantocmptlut[dec->cdef->data.cdef.ents[i].channo],
+			unsigned compno = dec->chantocmptlut[channo];
+			if (compno >= jas_image_numcmpts(dec->image)) {
+				jas_logerrorf(
+				  "error: invalid component reference in CDEF box (%d)\n",
+				  compno);
+				goto error;
+			}
+			jas_image_setcmpttype(dec->image, compno,
 			  jp2_getct(jas_image_clrspc(dec->image),
-			  dec->cdef->data.cdef.ents[i].type, dec->cdef->data.cdef.ents[i].assoc));
+			  dec->cdef->data.cdef.ents[i].type,
+			  dec->cdef->data.cdef.ents[i].assoc));
 		}
 	} else {
 		for (i = 0; i < dec->numchans; ++i) {
-			jas_image_setcmpttype(dec->image, dec->chantocmptlut[i],
+			unsigned compno = dec->chantocmptlut[i];
+			if (compno >= jas_image_numcmpts(dec->image)) {
+				jas_logerrorf(
+				  "error: invalid component reference (%d)\n", compno);
+				goto error;
+			}
+			jas_image_setcmpttype(dec->image, compno,
 			  jp2_getct(jas_image_clrspc(dec->image), 0, i + 1));
 		}
 	}
@@ -415,12 +497,9 @@ jas_image_t *jp2_decode(jas_stream_t *in, char *optstr)
 
 	/* Ensure that some components survived. */
 	if (!jas_image_numcmpts(dec->image)) {
-		jas_eprintf("error: no components\n");
+		jas_logerrorf("error: no components\n");
 		goto error;
 	}
-#if 0
-jas_eprintf("no of components is %d\n", jas_image_numcmpts(dec->image));
-#endif
 
 	/* Prevent the image from being destroyed later. */
 	image = dec->image;
@@ -431,6 +510,9 @@ jas_eprintf("no of components is %d\n", jas_image_numcmpts(dec->image));
 	return image;
 
 error:
+	if (image) {
+		jas_image_destroy(image);
+	}
 	if (box) {
 		jp2_box_destroy(box);
 	}
@@ -440,41 +522,25 @@ error:
 	return 0;
 }
 
+#define JP2_VALIDATE_LEN JAS_MIN(JP2_JP_LEN + 4, JAS_STREAM_MAXPUTBACK)
 int jp2_validate(jas_stream_t *in)
 {
-	char buf[JP2_VALIDATELEN];
-	int i;
-	int n;
-#if 0
-	jas_stream_t *tmpstream;
-	jp2_box_t *box;
-#endif
-
-	assert(JAS_STREAM_MAXPUTBACK >= JP2_VALIDATELEN);
+	unsigned char buf[JP2_VALIDATE_LEN];
 
 	/* Read the validation data (i.e., the data used for detecting
 	  the format). */
-	if ((n = jas_stream_read(in, buf, JP2_VALIDATELEN)) < 0) {
-		return -1;
-	}
-
-	/* Put the validation data back onto the stream, so that the
-	  stream position will not be changed. */
-	for (i = n - 1; i >= 0; --i) {
-		if (jas_stream_ungetc(in, buf[i]) == EOF) {
-			return -1;
-		}
-	}
-
-	/* Did we read enough data? */
-	if (n < JP2_VALIDATELEN) {
+	assert(sizeof(buf) <= JAS_STREAM_MAXPUTBACK);
+	if (jas_stream_peek(in, buf, sizeof(buf)) != sizeof(buf)) {
 		return -1;
 	}
 
 	/* Is the box type correct? */
-	if (((buf[4] << 24) | (buf[5] << 16) | (buf[6] << 8) | buf[7]) !=
-	  JP2_BOX_JP)
-	{
+	assert(JP2_VALIDATE_LEN >= 8);
+	if (((JAS_CAST(uint_least32_t, buf[4]) << 24) |
+	  (JAS_CAST(uint_least32_t, buf[5]) << 16) |
+	  (JAS_CAST(uint_least32_t, buf[6] << 8)) |
+	  (JAS_CAST(uint_least32_t, buf[7]))) !=
+	  JP2_BOX_JP) {
 		return -1;
 	}
 
@@ -539,38 +605,30 @@ static int jp2_getct(int colorspace, int type, int assoc)
 			switch (assoc) {
 			case JP2_CDEF_RGB_R:
 				return JAS_IMAGE_CT_COLOR(JAS_CLRSPC_CHANIND_RGB_R);
-				break;
 			case JP2_CDEF_RGB_G:
 				return JAS_IMAGE_CT_COLOR(JAS_CLRSPC_CHANIND_RGB_G);
-				break;
 			case JP2_CDEF_RGB_B:
 				return JAS_IMAGE_CT_COLOR(JAS_CLRSPC_CHANIND_RGB_B);
-				break;
 			}
 			break;
 		case JAS_CLRSPC_FAM_YCBCR:
 			switch (assoc) {
 			case JP2_CDEF_YCBCR_Y:
 				return JAS_IMAGE_CT_COLOR(JAS_CLRSPC_CHANIND_YCBCR_Y);
-				break;
 			case JP2_CDEF_YCBCR_CB:
 				return JAS_IMAGE_CT_COLOR(JAS_CLRSPC_CHANIND_YCBCR_CB);
-				break;
 			case JP2_CDEF_YCBCR_CR:
 				return JAS_IMAGE_CT_COLOR(JAS_CLRSPC_CHANIND_YCBCR_CR);
-				break;
 			}
 			break;
 		case JAS_CLRSPC_FAM_GRAY:
 			switch (assoc) {
 			case JP2_CDEF_GRAY_Y:
 				return JAS_IMAGE_CT_COLOR(JAS_CLRSPC_CHANIND_GRAY_Y);
-				break;
 			}
 			break;
 		default:
 			return JAS_IMAGE_CT_COLOR(assoc - 1);
-			break;
 		}
 	}
 	return JAS_IMAGE_CT_UNKNOWN;
@@ -582,13 +640,10 @@ static int jp2_getcs(jp2_colr_t *colr)
 		switch (colr->csid) {
 		case JP2_COLR_SRGB:
 			return JAS_CLRSPC_SRGB;
-			break;
 		case JP2_COLR_SYCC:
 			return JAS_CLRSPC_SYCBCR;
-			break;
 		case JP2_COLR_SGRAY:
 			return JAS_CLRSPC_SGRAY;
-			break;
 		}
 	}
 	return JAS_CLRSPC_UNKNOWN;
@@ -599,13 +654,10 @@ static int fromiccpcs(int cs)
 	switch (cs) {
 	case ICC_CS_RGB:
 		return JAS_CLRSPC_GENRGB;
-		break;
 	case ICC_CS_YCBCR:
 		return JAS_CLRSPC_GENYCBCR;
-		break;
 	case ICC_CS_GRAY:
 		return JAS_CLRSPC_GENGRAY;
-		break;
 	}
 	return JAS_CLRSPC_UNKNOWN;
 }
